/*
 *	elf_analy.cc
 */

/*	Copyright (C) 1999, 2000, 2001 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "analy.h"
#include "analy_alpha.h"
#include "analy_names.h"
#include "analy_x86.h"
#include "codeanaly.h"
#include "global.h"
#include "elf_analy.h"

#include "htctrl.h"
#include "htdebug.h"
#include "htiobox.h"
#include "htelf.h"
#include "htstring.h"
#include "pestruct.h"
#include "x86asm.h"

extern "C" {
#include "demangle.h"
}

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

char		elf_buffer[1024];

/*
 *
 */
 
void	elf_analyser::init(ht_elf_shared_data *Elf_shared, ht_stream *File)
{
	elf_shared = Elf_shared;
	file = File;

	ADDR entry=elf_shared->header.e_entry;

	validarea = new area();
	validarea->init();

	analyser::init();

	/////////////

	set_addr_tree_optimize_threshold(100);
	set_label_tree_optimize_threshold(100);

	// give all sections a descriptive comment:

	ELF_SECTION_HEADER *s=elf_shared->sheaders.sheaders;
	char blub[100];
	for (UINT i=0; i<elf_shared->sheaders.count; i++) {
		ADDR secaddr = s->sh_addr;
		if (validaddr(secaddr, scvalid)) {
			sprintf(blub, ";  section %d <%s>", i+1, getaddrsectionname(secaddr));
			addcomment(secaddr, 0, " ");
			addcomment(secaddr, 0, ";******************************************************************");
			addcomment(secaddr, 0, ht_strdup(blub));
			sprintf(blub, ";  virtual address  %08x  virtual size   %08x", s->sh_addr, s->sh_size);
			addcomment(secaddr, 0, ht_strdup(blub));
			sprintf(blub, ";  file offset      %08x  file size      %08x", s->sh_offset, s->sh_size);
			addcomment(secaddr, 0, ht_strdup(blub));
			addcomment(secaddr, 0, ";******************************************************************");

			// mark end of sections
			sprintf(blub, ";  end of section <%s>", getaddrsectionname(secaddr));
			ADDR secend_addr = secaddr + s->sh_size;
			newaddr(secend_addr)->flags |= AF_FUNCTION_END;
			addcomment(secend_addr, 0, "");
			addcomment(secend_addr, 0, ";******************************************************************");
			addcomment(secend_addr, 0, ht_strdup(blub));
			addcomment(secend_addr, 0, ";******************************************************************");

			validarea->add(secaddr, secend_addr);
		}

		s++;
	}

/* symbols */
	for (UINT i=1; i<elf_shared->sheaders.count; i++) {
		if ((elf_shared->sheaders.sheaders[i].sh_type==ELF_SHT_SYMTAB) || (elf_shared->sheaders.sheaders[i].sh_type==ELF_SHT_DYNSYM)) {
			init_insert_symbols(i);
		}
	}

	set_addr_tree_optimize_threshold(1000);
	set_label_tree_optimize_threshold(1000);

	addcomment(entry, 0, " ");
	addcomment(entry, 0, ";****************************");
	switch (elf_shared->header.e_type) {
		case ELF_ET_DYN:
			addcomment(entry, 0, ";  dynamic executable entry point");
			break;
		case ELF_ET_EXEC:
			addcomment(entry, 0, ";  executable entry point");
			break;
		default:
			addcomment(entry, 0, ";  entry point");
	}
	addcomment(entry, 0, ";****************************");
	assignlabel(entry, ht_strdup("entrypoint"), label_func);
}

/*
 *
 */
void	elf_analyser::init_insert_symbols(int shidx)
{
	STREAMOFS h=elf_shared->sheaders.sheaders[shidx].sh_offset;
	STREAMOFS sto=elf_shared->sheaders.sheaders[elf_shared->sheaders.sheaders[shidx].sh_link].sh_offset;
	UINT symnum=elf_shared->sheaders.sheaders[shidx].sh_size / sizeof (ELF_SYMBOL);

	int *entropy = random_permutation(symnum);
	
	for (UINT i=0; i<symnum; i++) {
		ELF_SYMBOL sym;
		file->seek(h+entropy[i]*sizeof (ELF_SYMBOL));
//		file->seek(h+i*sizeof (ELF_SYMBOL));
		file->read(&sym, sizeof sym);
		file->seek(sto+sym.st_name);
		char *name=getstrz(file);

		switch (sym.st_shndx) {
			case ELF_SHN_UNDEF:
				break;
			case ELF_SHN_ABS:
				break;
			case ELF_SHN_COMMON:
				break;
			default: {
				// sym.st_shndx
				break;
			}
		}

		char *bind="?";
		switch (ELF32_ST_BIND(sym.st_info)) {
			case ELF_STB_LOCAL:
				bind="local";
				break;
			case ELF_STB_GLOBAL:
				bind="global";
				break;
			case ELF_STB_WEAK:
				bind="weak";
				break;
		}
		
		switch (ELF32_ST_TYPE(sym.st_info)) {
			case ELF_STT_FUNC: {
				char label[1024];
				sprintf(label, "%s", name);
				if (!findlabel(label)) {
					ADDR address=sym.st_value;

					char *demangled = cplus_demangle(label, DMGL_PARAMS | DMGL_ANSI);
					
					sprintf(elf_buffer, "; function %s (%s)", (demangled) ? demangled : label, bind);

					if (demangled) free(demangled);
			
					addcomment(address, 0, " ");
					addcomment(address, 0, ";********************************************************");
					addcomment(address, 0, ht_strdup(elf_buffer));
					addcomment(address, 0, ";********************************************************");
					pushaddr(address, address);
					assignlabel(address, ht_strdup(label), label_func);
				}
				break;
			}
			case ELF_STT_OBJECT: {
				char label[1024];
				sprintf(label, "%s", name);
				if (!findlabel(label)) {
					ADDR address=sym.st_value;
		
					char *demangled = cplus_demangle(label, DMGL_PARAMS | DMGL_ANSI);
					
					sprintf(elf_buffer, "; data object %s, size %d (%s)", (demangled) ? demangled : label, sym.st_size, bind);

					if (demangled) free(demangled);
			
					addcomment(address, 0, " ");
					addcomment(address, 0, ";********************************************************");
					addcomment(address, 0, ht_strdup(elf_buffer));
					addcomment(address, 0, ";********************************************************");
					assignlabel(address, ht_strdup(label), label_data);
				}
				break;
			}
			case ELF_STT_SECTION:
			case ELF_STT_FILE:
				break;
		}
		delete name;
	}
	free(entropy);
}

/*
 *
 */
int	elf_analyser::load(ht_stream_filter *f)
{
	/*
	ht_elf_shared_data 	*elf_shared;
	ht_stream 		*file;
	area				*validarea;
	*/

	return analyser::load(f);
}

/*
 *
 */
void	elf_analyser::done()
{
	validarea->done();
	delete validarea;
	analyser::done();
}

/*
 *
 */
void *elf_analyser::bufptr(ADDR Addr)
{
	FILEADDR ofs = fileaddr(Addr);
	assert(ofs != INVALID_FILE_OFS);
	file->seek(ofs);
	file->read(elf_buffer, 16);
	return (void*)elf_buffer;
}

/*
 *
 */
assembler *elf_analyser::create_assembler()
{
	switch (elf_shared->header.e_machine) {
		case ELF_EM_386:
			assembler *a = new x86asm(X86_OPSIZE32, X86_ADDRSIZE32);
			a->init();
			return a;
	}
	return NULL;
}

byte elf_digit(char c)
{
	if ((c>='0') && (c<='9')) return (c-'0');
	if ((c>='a') && (c<='f')) return (c-'a'+10);
	if ((c>='A') && (c<='F')) return (c-'A'+10);
	return 42;
}

/*
 *
 */
void	elf_analyser::examineopcode(OPCODE *opcode)
{
	/* _bad_ code, doesn't even work */
	char *opcode_str = disasm->str(opcode, 0);

	ADDR	offset = 0;
	if (!opcode_str) return;
	while (*opcode_str) {
		if (elf_digit(*opcode_str) < 16) {
			do {
				offset *= 16;
				offset += elf_digit(*opcode_str++);
			} while ((*opcode_str) && (elf_digit(*opcode_str) < 16));
			taccess a;
			a.indexed = false;
			a.type = acread;
			if (validaddr(offset, scvalid)) {
				dataaccess(offset, a);
				addxref(offset, addr, xrefoffset);
			}
			offset = 0;
			if (!*opcode_str) return;
		}
		opcode_str++;
	}
}

/*
 *
 */
FILEADDR elf_analyser::fileaddr(ADDR Addr)
{
	if (validaddr(Addr, scinitialized)) {
		dword ofs;
		if (!elf_addr_to_ofs(&elf_shared->sheaders, Addr, &ofs)) return INVALID_FILE_OFS;
		return ofs;
	} else {
		return INVALID_FILE_OFS;
	}
}

/*
 *
 */
char elf_sectionname[33];

char *elf_analyser::getaddrsectionname(ADDR Addr)
{
	elf_section_headers *sections=&elf_shared->sheaders;
	int i;
	if (!elf_addr_to_section(sections, Addr, &i)) return NULL;
	strncpy(elf_sectionname, elf_shared->shnames[i], 32);
	elf_sectionname[32]=0;
	return elf_sectionname;
}

/*
 *
 */
char	*elf_analyser::getname()
{
	return file->get_name();
}

/*
 *
 */
void elf_analyser::init_codeanalyser()
{
	analyser::init_codeanalyser();
	code->loaddefs("analyser/sign.def");
}

/*
 *
 */
void elf_analyser::init_unasm()
{
	dprintf("elf_analy: ");
	switch (elf_shared->header.e_machine) {
		case ELF_EM_386: // Intel
			dprintf("initing analy_x86_disassembler\n");
			analy_disasm = new analy_x86_disassembler();
			analy_disasm->init(this);
			break;
		default:
			dprintf("no apropriate disassembler for machine %04x\n", elf_shared->header.e_machine);
			warnbox("No disassembler for unknown machine type %04x!", elf_shared->header.e_machine);
	}
}

/*
 *
 */
void elf_analyser::log(char *msg)
{
	/*
	 *	log() does to much traffic so dont log
	 *   perhaps we reactivate this later
	 *
	 */
/*	LOG(msg);*/
}

/*
 *
 */
ADDR elf_analyser::nextvalid(ADDR Addr)
{
	return INVALID_ADDR; //validarea->findnext(Addr);
	// FIXME (hack while validarea isnt active):
//   taddr *a = enum_addrs(Addr);
//   return (a)?a->addr:INVALID_ADDR;
}

/*
 *
 */
void elf_analyser::store(ht_stream_filter *f)
{
	/*
	ht_elf_shared_data 	*elf_shared;
	ht_stream 		*file;
	area				*validarea;
	*/
	f->put_object(validarea, "validarea");
	analyser::store(f);
}

/*
 *
 */
int	elf_analyser::query_config(int mode)
{
	switch (mode) {
		case Q_DO_ANALYSIS:
		case Q_ENGAGE_CODE_ANALYSER:
		case Q_ENGAGE_DATA_ANALYSER:
			return true;
		default:
			return 0;
	}
}

/*
 *
 */
ADDR elf_analyser::vaddr(FILEADDR fileaddr)
{
	ADDR a;
	if (elf_ofs_to_addr(&elf_shared->sheaders, fileaddr, &a)) {
		return a;
	} else {
		return INVALID_ADDR;
	}
}

/*
 *
 */
bool elf_analyser::validaddr(ADDR Addr, tsectype action)
{
	elf_section_headers *sections=&elf_shared->sheaders;
	int sec;
	if (!elf_addr_to_section(sections, Addr, &sec)) return false;
	ELF_SECTION_HEADER *s=sections->sheaders+sec;
	switch (action) {
		case scvalid:
			return true;
		case scread:
			return true;
		case scwrite:
		case screadwrite:
			return s->sh_flags & ELF_SHF_WRITE;
		case sccode:
			return (s->sh_flags & ELF_SHF_EXECINSTR) && (s->sh_type==ELF_SHT_PROGBITS);
		case scinitialized:
			return s->sh_type==ELF_SHT_PROGBITS;
	}
	return false;
}


