/*
 *	htanaly.cpp
 */

/*	Copyright (C) 1999, 2000, 2001 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "analy.h"
#include "analy_names.h"
#include "global.h"
#include "htanaly.h"
#include "htapp.h"
#include "htdialog.h"
#include "hthist.h"
#include "htidle.h"
#include "htiobox.h"
#include "htkeyb.h"
#include "htstring.h"
#include "httag.h"
#include "httree.h"
#include "malloc.h"
#include "tools.h"

#include <stdlib.h>
#include <string.h>

//test
#include "out_html.h"

#include "x86dis.h"

/*
 *	symbolbox
 */

void symbolbox::init(bounds *b, analyser *Analy)
{
	analy = Analy;
	ht_listbox::init(b);
	str = (char *)malloc(1024); // Buffer overflow...
}

void symbolbox::done()
{
	ht_listbox::done();
}

int  symbolbox::calc_count()
{
	return analy->getlabelcount();
}

int  symbolbox::cursor_adjust()
{
	return 18;
}

void *symbolbox::getfirst()
{
	return analy->enum_labels(NULL);
}

void *symbolbox::getlast()
{
	return analy->enum_labels_back(NULL);
}

void *symbolbox::getnext(void *entry)
{
	if (!entry) return NULL;
	return analy->enum_labels(((tlabel *)entry)->name);
}

void *symbolbox::getnth(int n)
{
	// slow!!!!!!!!!!
	// n==1 =^= 1th entry
	tlabel *e = NULL;
	for (int i=0; i<n; i++) {
		e = analy->enum_labels(((tlabel *)e)->name);
		if (!e) break;
	}
	return e;
}

void *symbolbox::getprev(void *entry)
{
	if (!entry) return NULL;
	return analy->enum_labels_back(((tlabel *)entry)->name);
}

char *symbolbox::getstr(int col, void *entry)
{
	if (!entry) return NULL;
	tlabel *l = ((tlabel *)entry);
	switch (col) {
		case 0:
			sprintf(str, "%8x", l->addr->addr);
			break;
		case 1:
			strcpy(str, label_type_short(l->type));
			break;
		case 2:
			strcpy(str, l->name);
			break;
	}
	return str;
}

bool symbolbox::idle()
{
	if ((idle_count % 500)==0) {
		update();
		redraw();
		idle_count = 1;
		return 1;
	}
	idle_count++;
	return 0;
}

int	symbolbox::num_cols()
{
	return 3;
}

void *symbolbox::quickfind(char *s)
{
	tlabel *tmp = analy->enum_labels(s);
	if (tmp) {
		int slen = strlen(s);
		int tlen = strlen(tmp->name);
		if (slen > tlen) return NULL;
		if (strncmp(tmp->name, s, slen)==0) return tmp;
		return NULL;
	} else {
		return NULL;
	}
}


char	*symbolbox::quickfind_completition(char *s)
{
	if (analy->findlabel(s)) {
		return ht_strdup(s);
	}
	tlabel *tmp = analy->enum_labels(s);
	if (!tmp) {
		return ht_strdup(tmp->name);
	}
	tlabel *tmp2 = analy->enum_labels(tmp->name);
	if (!tmp2) {
		return ht_strdup(tmp->name);
	}
	int slen = strlen(s);
	if (!ht_strncmp(tmp->name, tmp2->name, slen)==0) {
		return ht_strdup(tmp->name);
	}
	char *res = (char *)malloc(1024);
	strcpy(res, tmp->name);
	while (tmp2 && (ht_strncmp(tmp2->name, s, slen)==0)) {
//		fprintf(stdout, "while(%s, %s, %d)\n", tmp2->name, s, slen);
		int a = strccomm(res, tmp2->name);
		res[a] = 0;
		tmp2 = analy->enum_labels(tmp2->name);
	}
	return res;
}


/////////////////////////////////////////////////////////////////////////////

/*
 *
 */
void analy_infoline::init(bounds *b, ht_aviewer *A, char *Format)
{
	ht_statictext::init(b, 0, align_left);
	VIEW_DEBUG_NAME("analy_infoline");
	analy = A;
	displayformat = ht_strdup(Format);
	s = (char *)malloc(1024);
}

void analy_infoline::done()
{
	free(displayformat);
	free(s);
	ht_statictext::done();
}

char *analy_infoline::gettext()
{
	if (valid()) {
		char *sec = analy->analy->getaddrsectionname(addr);
		if (fofs!=INVALID_FILE_OFS) {
			taddr *a = analy->analy->findaddrfunc(addr);
			char *func = (a) ? ((a->label) ? a->label->name : NULL): NULL;
			sprintf(s, displayformat, (sec) ? sec : "", fofs,  analy->analy->getdisasmstr(addr), (func)?func:"");
		} else {
			if (!sec) {
				strcpy(s, "[not in file]");
			} else {
				sprintf(s, "<%s> [not in file]", sec);
			}
		}
		return s;
	} else {
		return "<no analyser>";
	}
}

void analy_infoline::update(ADDR cursor_addr, ADDR ecursor_addr)
{
	if (valid()) {
		if (ecursor_addr) {
			fofs = ecursor_addr;
		} else {
			fofs = analy->analy->fileaddr(cursor_addr);
		}
		addr = cursor_addr;
	} else {
		fofs = INVALID_FILE_OFS;
		addr = INVALID_ADDR;
	}
	dirtyview();
	redraw();
}

bool analy_infoline::valid()
{
	return ((analy)&&(analy->analy));
}

void ht_aviewer::init(bounds *b, char *desc, int caps, ht_stream *stream, ht_format_group *format_group, analyser *Analy)
{
	analy=Analy;
	analy->setdisplaymode(ANALY_TRANSLATE_SYMBOLS+ANALY_COLLAPSE_XREFS, 0);
	ht_uformat_view::init(b, desc, caps, stream, format_group);
	infoline=NULL;
	idle_count=0;
	last_active=1;
	register_idle_object(this);
}

void ht_aviewer::done()
{
	unregister_idle_object(this);
	ht_uformat_view::done();
	analy->done();
	delete analy;
}

void ht_aviewer::attach_infoline(analy_infoline *V)
{
	infoline = V;
}

bool ht_aviewer::address_to_offset(fmt_vaddress addr, STREAMOFS *ofs)
{
	STREAMOFS o=analy->fileaddr(addr);
	if (o!=INVALID_FILE_OFS) {
		*ofs=o;
		return 1;
	}
	return 0;
}

char *ht_aviewer::func(UINT i, bool execute)
{
	switch (i) {
		case 6: {
			if (execute) {
				bounds b;
				BOUNDS_ASSIGN(b, 10, 5, 60, 15);
				ht_dialog *dialog = new ht_dialog();
				dialog->init(&b, "symbols", FS_KILLER | FS_TITLE | FS_MOVE);
				/* pull down */
				BOUNDS_ASSIGN(b, 30, 0, 20, 1);
				ht_listpopup *lp = new ht_listpopup();
				lp->init(&b);
				lp->insertstring("show all");
				lp->insertstring("only functions");
				lp->insertstring("only labels");
				lp->insertstring("only offsets (code)");
				lp->insertstring("only data");
//				dialog->insert(lp);
				/* text */
				BOUNDS_ASSIGN(b, 1, 0, 30, 1);
				ht_statictext *text = new ht_statictext();
				text->init(&b, " Address   Type   Name", 0);
				dialog->insert(text);
				/* list */
				BOUNDS_ASSIGN(b, 1, 1, 56, 12);
				symbolbox *sym=new symbolbox();
				sym->init(&b, analy);
				dialog->insert(sym);
				register_idle_object(sym);
				int r = dialog->run(0);
				unregister_idle_object(sym);
				if (r) {
					// goto selected symbol
					ht_listbox_data d;
					sym->getdata(&d);
					fmt_vaddress vaddr;
					if (string_to_address(((tlabel *)(d.cursor))->name, &vaddr)) {
						goto_address(vaddr);
					}
				}
				dialog->done();
				delete dialog;
			}
			return "symbols";
		}
		default:
			return ht_uformat_view::func(i, execute);
	}
	return 0;
}

void ht_aviewer::handlemsg(htmsg *msg)
{
	char str[1024];
	switch (msg->msg) {
		case msg_draw:
			if (infoline) {
				FILEADDR a;
				infoline->update(cursor_id1, (get_current_offset(&a)) ? a : 0);
			}
			break;
		case msg_keypressed:
			switch (msg->data1.integer) {
				case K_Control_A: {
					assembler *a = analy->create_assembler();
					if (!a) {
						// select assembler
						infobox("select assembler");
						clearmsg(msg);
						return;
					}
					char instr[257];
					instr[0] = 0;
					fmt_vaddress current_address;
					get_current_address(&current_address);
					strcpy(instr, analy->getdisasmstr(current_address));
					asm_insn *insn = a->alloc_insn();
					asm_code *ac = NULL;
					while (inputbox(a->get_name(), "instruction:", instr, 255, HISTATOM_ASSEMBLER)) {
						if (a->prepare_str(insn, instr)) {
							ac = a->encode(insn, 0, analy->mapaddr(current_address));
							if (ac) {
	//                              	infobox("size: %d", ac->size);
								break;
							} else {
								errorbox("%s: %s", a->get_name(), a->get_error_msg());
							}
						} else {
							errorbox("%s: %s", a->get_name(), a->get_error_msg());
						}
					}
					if (ac) {
						bool ok=true;
						asm_code *chosen_ac = ac;
						if (ac->next) {
							// choose from list if ambigous
							bounds b;
							BOUNDS_ASSIGN(b, 10, 5, 60, 15);
							ht_dialog *dialog = new ht_dialog();
							dialog->init(&b, "choose opcode", FS_KILLER | FS_TITLE | FS_MOVE);
							BOUNDS_ASSIGN(b, 1, 0, 56, 1);
							ht_statictext *text = new ht_statictext();
							text->init(&b, " opcode (& disassembly)", 0);
							dialog->insert(text);
							BOUNDS_ASSIGN(b, 1, 1, 56, 12);
							ht_text_listbox *list=new ht_text_listbox();
							list->init(&b, 2, 0);
							asm_code *ac2 = ac;
							UINT aci=0;
							while (ac2) {
								char s[1024];
								s[0] = 0;
								char *t = s;
								for (unsigned int i=0; i<ac2->size; i++) {
									t += sprintf(t, "%02x ", ac2->data[i]);
								}
								char *d;
								if (analy->disasm) {
									OPCODE *o=analy->disasm->decode((byte *)ac2->data, ac2->size, analy->mapaddr(current_address));
									d = analy->disasm->strf(o, X86DIS_STYLE_HEX_NOZEROPAD+X86DIS_STYLE_HEX_ASMSTYLE, DISASM_STRF_SMALL_FORMAT);
								} else {
									d = "<no disassembler>";
								}
								list->insert_str(aci, s, d);
								ac2 = ac2->next;
								aci++;
							}
							list->update();
							dialog->insert(list);
							int r = dialog->run(0);
							ok = r;
							if (r) {
								ht_listbox_data d;
								list->getdata(&d);
								ht_text_listbox_item *i=(ht_text_listbox_item *)d.cursor;
								asm_code *ac3 = ac;
								int ac3i=0;
								while (ac3) {
									if (ac3i==i->id) {
										chosen_ac=ac3;
										break;
									}
									ac3=ac3->next;
									ac3i++;
								}
							}
							dialog->done();
							delete dialog;
						}
						if (ok) {
							fmt_vaddress vaddr;
							if (get_current_address(&vaddr)) {
								sendmsg(cmd_edit_mode);
								if (edit) {
									vwrite(vaddr, chosen_ac->data, chosen_ac->size);
								}
							}
						}
					}
					free(insn);
					if (a) {
						a->done();
						delete a;
					}
					clearmsg(msg);
					return;
				}
				case K_Control_B: {
					clearmsg(msg);
					return;
				}
				case K_Control_F: {
					taddr *a = analy->findaddrfunc(cursor_id1);
					if (a) {
						goto_address(a->addr, this);
					} else {
						errorbox("Address %08lx doesn't belong to a function.", cursor_id1);
					}
					clearmsg(msg);
					return;
				}
				case K_Control_L: {
					taddr *a = analy->findaddrlabel(cursor_id1);
					if (a) {
						goto_address(a->addr, this);
					} else {
						errorbox("Address %08lx doesn't belong to a label.", cursor_id1);
					}
					clearmsg(msg);
					return;
				}
				case K_Control_O: {
					ht_file_stream *s = new ht_file_stream();
					s->init("__test.htm", SAM_CREATE+SAM_WRITE);
					generate_html_output(analy, s, 0x406400, 0x406400+0x1000);
					s->done();
					delete s;
					clearmsg(msg);
					return;
				}
				case 'c':
					if (!edit) {
						analy->continueanalysisat(cursor_id1);
						clearmsg(msg);
						return;
					}
					break;
				case '#': {
					tcomment *c = analy->getcomments(cursor_id1);
					if (c) {
						tcomment *d = c;
						int size = 3;
						while (d) {
							size += strlen(d->text)+1;
							d = d->next;
						}
						char *text = (char *)malloc(size);
						char *t = text;
						t += sprintf(t, "\el");
						d = c;
						while (d) {
							t += sprintf(t, "%s\n", d->text);
							d = d->next;
						}
						*t = 0;
						infobox_c(text);
						free(text);
					} else {
						errorbox("no comments.");
					}
					clearmsg(msg);
					return;
				}
				case 'n': {
					dword addr=cursor_id1;
					if (analy->validaddr(addr, scvalid)) {
						char n[255];
						tlabel *l = analy->getaddrlabel(addr);
						if (l) strcpy(n, l->name); else n[0] = 0;
						sprintf(str, "name for address %08x", addr);
						if (inputbox(str, "label name:", n, 255)) {
							if (n[0]) {
								if (valid_name(n)) {
									char *n2 = ht_strdup(n);
									if (!analy->assignlabel(addr, n2, (l)?l->type:label_loc)) {
										l = analy->findlabel(n);
										errorbox("Label `%s' already exists at address %08x!", n, l->addr->addr);
										free(n2);
									}
								} else {
									errorbox("`%s' is an invalid label name.", n);
								}
							} else {
								if (l) {
									// delete label if applicable
									if (confirmbox("Really delete label `%s' at address %08x?", l->name, addr)) {
										analy->deletelabel(addr);
									}
								}
							}
						}
					} else {
						errorbox("Address %08lx is invalid!", addr);
					}
					clearmsg(msg);
					return;
				}
				case 'x': {
					show_xrefs(cursor_id1);
					clearmsg(msg);
					return;
				}
				case 'f': {
					if (!edit) {
						ADDR *ptr = (ADDR *)analy->bufptr(cursor_id1);
						ADDR p = *ptr;
						if (analy->validaddr(p, scvalid)) {
							goto_address(p, this);
						} else {
							errorbox("Follow: address %08lx is invalid!", p);
						}
						clearmsg(msg);
						return;
					}
					break;
				}
				case 'F': {
					clearmsg(msg);
					return;
				}
			}
			break;
		case cmd_edit_mode: {
			analy->setdisplaymode(ANALY_EDIT_BYTES, 0);
			ht_uformat_view::handlemsg(msg);
			if (!edit) analy->setdisplaymode(0, ANALY_EDIT_BYTES);
			clearmsg(msg);
			return;
		}
		case cmd_view_mode: {
			analy->setdisplaymode(0, ANALY_EDIT_BYTES);
			ht_uformat_view::handlemsg(msg);
			if (edit) analy->setdisplaymode(ANALY_EDIT_BYTES, 0);
			clearmsg(msg);
			return;
		}
	}
	ht_uformat_view::handlemsg(msg);
}

bool ht_aviewer::idle()
{
	last_active=analy->active;
	analy->continueanalysis();
	if (last_active && !analy->active) {
		LOG("%s: analyser finished after %d ops.", analy->getname(), analy->ops_parsed);
		dirtyview();
		app->sendmsg(msg_draw, 0);
	}
	if (analy->active) {
		idle_count++;
		if (idle_count%50==0) {
			dirtyview();
			app->sendmsg(msg_draw, 0);
		}
	}
	return analy->active;
}

bool ht_aviewer::offset_to_address(STREAMOFS ofs, fmt_vaddress *addr)
{
	*addr = analy->vaddr(ofs);
	return (*addr != INVALID_ADDR);
}

int ht_aviewer::ref_sel(ID id_low, ID id_high)
{
	switch (id_high) {
		case 0:
			return goto_address(id_low, this);
		case 1:
			show_xrefs((ADDR) id_low);
			return 0;
	}
	return 0;
}

void ht_aviewer::show_xrefs(ADDR Addr)
{
	txref *x = analy->getxrefs(Addr);
	if (x) {
		bounds b;
		BOUNDS_ASSIGN(b, 10, 5, 60, 15);
		char str[100];
		sprintf(str, "xrefs of address %08x", Addr);
		ht_dialog *dialog = new ht_dialog();
		dialog->init(&b, str, FS_KILLER | FS_TITLE | FS_MOVE);
		BOUNDS_ASSIGN(b, 1, 0, 56, 10);
		ht_statictext *text = new ht_statictext();
		text->init(&b, " xref to   type     from function", 0);
		dialog->insert(text);
		BOUNDS_ASSIGN(b, 1, 1, 56, 10);
		ht_text_listbox *list;
		NEW_OBJECT(list, ht_text_listbox, &b, 3, 2);
		while (x) {
			sprintf(str, "%8x", x->addr);
			taddr *a = analy->findaddrfunc(x->addr);
			char *func = (a) ? ((a->label) ? a->label->name : NULL): NULL;
			list->insert_str(x->addr, str, xref_type(x->type), (func)?func:"?");
			x = x->next;
		}
		list->update();
		dialog->insert(list);
		int r = dialog->run(0);
		if (r) {
			ht_listbox_data data;
			list->getdata(&data);
			goto_address(((ht_text_listbox_item *)data.cursor)->id, this);
		}
		dialog->done();
		delete dialog;
	} else {
		errorbox("No xrefs for address %08lx!", Addr);
	}
}

int ht_aviewer_symbol_to_addr(ht_aviewer *aviewer, char **s, dword *v)
{
	if (**s == '@') {
		(*s)++;
		if (bnstr(s, v, 10)) {
			if (!aviewer->offset_to_address(*v, v)) {
				return 0;
			}
			return 1;
		}
		// invalid number after @
		return 0;
	} else {
		char *k=ht_strdup(*s);
		char *t=k;
		while (!strchr("+-*/ \t", *t) && *t) t++;
		char temp=*t;
		*t=0;
		if ((*k == '$') && (k[1] == 0)) {
			*t=temp;
			*s+= t-k;
			*v = aviewer->cursor_id1;
			free(k);
			return 1;
		}
		tlabel *l = aviewer->analy->findlabel(k);
		*t=temp;
		if (l) {
			// Label
			*s+=t-k;
			*v=l->addr->addr;
			free(k);
			return 1;
/*		} else if (**s=='.') {
			// RVA
			if (bnstr(s, v, 16)) {
				return 1;
			}*/
		}
		free(k);
	}
	return 0;
}

bool ht_aviewer::string_to_address(char *string, fmt_vaddress *vaddr)
{
	ADDR a;
	int ev=eval_int_expr_custom(string, &a, this, ht_aviewer_symbol_to_addr);
	if (ev) {
		if (analy->validaddr(a, scvalid)) {
			*vaddr=a;
			return 1;
		} else {
			sprintf(globalerror, "address %08x is invalid", a);
		}
	} else {
		strcpy(globalerror, eval_int_errstr);
	}
	return 0;
}

/*
 *	CLASS hts_analy
 */

void hts_analy::init(ht_stream *file, analyser *A, ADDR Lowestaddress, ADDR Highestaddress)
{
	hts_sub::init(file);
	analy = A;
	lowestaddress=Lowestaddress;
	highestaddress=Highestaddress-1;
}

void hts_analy::done()
{
	hts_sub::done();
}

bool	hts_analy::convert_addr_to_id(fmt_vaddress addr, ID *id1, ID *id2)
{
	ADDR a=addr; /* FIXME: */
	if (analy->validaddr(a, scvalid)) {
		*id1=addr;
		*id2=0;
		return true;
	}
	return false;
}

bool	hts_analy::convert_id_to_addr(ID id1, ID id2, fmt_vaddress *addr)
{
	*addr=id1;
	return 1;
}

void	hts_analy::first_line_id(ID *id1, ID *id2)
{
	*id1=lowestaddress;
	*id2=0;
}

bool hts_analy::getline(char *line, ID id1, ID id2)
{
	int len;
	analy->getformattedline(id1, id2+1, line, &len);
	return 1;
}

void	hts_analy::last_line_id(ID *id1, ID *id2)
{
	*id1=highestaddress;
	*id2=0;
}

int	hts_analy::next_line_id(ID *id1, ID *id2, int n)
{
	int len;
	int res=0;
	while (n--) {
		if (analy->getformattedline_length(*id1, *id2+2, &len)) {
			(*id2)++;
		} else {
			analy->getformattedline_length(*id1, *id2+1, &len);
			if (*id1>=highestaddress) return res;
			*id1 += len;
			*id2 = 0;
		}
		res++;
	}
	return res;
}

int	hts_analy::prev_line_id(ID *id1, ID *id2, int n)
{
/* FIXME: interesting solution! */
	ADDR addrbuf[1024];
	int	linebuf[1024];
	int	i=0;
	int len;

/*
 *	24==1*2^3*3
 *   so it "supports" opcodes of length 1, 2, 3, 4, 6, 8 and 12
 *   (and of course 24)
 */
	ADDR	l=n*24;

/*
 *	120==24*5
 *   so 120 additionally "supports" opcodes of length 5, 10 and 15
 *   (and some more)
 */
	if (l<120) l=120;

	/* look backwards */
//     fprintf(stdout, "============= up ==============================\n");
//     fprintf(stdout, "%d x from %lx line %ld\n", n, *id1, *id2);
	ADDR tmpaddr = (*id1) - l;
	ADDR tmpaddr2;
	dword tmpline;
	first_line_id(&tmpaddr2, &tmpline);
	if ((l > (*id1)) || (tmpaddr < tmpaddr2)) tmpaddr = tmpaddr2;
	ADDR	nextaddr = tmpaddr;
	while (1) {
		if (tmpaddr >= (*id1)) {
			if ((tmpline >= (*id2)) || (tmpaddr > *id1)) break;
		}
		if (analy->getformattedline_length(tmpaddr, tmpline+1, &len)) {
			nextaddr += len;
			tmpline++;
		} else {
			tmpaddr = nextaddr;
			tmpline = 0;
			continue;
		}
//          fprintf(stdout, "%2d. %lx line %ld\n", i, tmpaddr, tmpline-1);
		addrbuf[i&(1023)] = tmpaddr;
		linebuf[i&(1023)] = tmpline-1;
		i++;
	}

	if (!i) return 0;

	if (i >= n) {
//     	fprintf(stdout, "result1: %lx line %d (%d)\n", addrbuf[i-n], linebuf[i-n], n);
		(*id1) = addrbuf[(i-n)&(1023)];
		(*id2) = linebuf[(i-n)&(1023)];
		return n;
	} else {
//     	fprintf(stdout, "result2: %lx line %d (%d)\n", addrbuf[0], linebuf[0], i);
		(*id1) = addrbuf[0];
		(*id2) = linebuf[0];
		return i;
	}
}

/*int hts_analy::search(ht_search *s, STREAMOFS start, STREAMOFS end, STREAMOFS *offset, dword *size)
{
	return SR_NOT_SUPPORTED;
}*/

