/* 
 *	HT Editor
 *	htelfphs.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "elfstruc.h"
#include "htatom.h"
#include "htelf.h"
#include "htelfphs.h"
#include "httag.h"
#include "formats.h"

hts_mask_ptable elfprogramheader[]=
{
	{"type",				STATICTAG_EDIT_DWORD("00000000")" ("STATICTAG_DESC_DWORD("00000000", ATOM_ELF_PH_TYPE_STR)")"},
	{"offset",			STATICTAG_EDIT_DWORD("00000004")},
	{"virtual address",		STATICTAG_EDIT_DWORD("00000008")},
	{"physical address",	STATICTAG_EDIT_DWORD("0000000c")},
	{"in file size",		STATICTAG_EDIT_DWORD("00000010")},
	{"in memory size",		STATICTAG_EDIT_DWORD("00000014")},
	{"flags",				STATICTAG_EDIT_DWORD("00000018")" "STATICTAG_FLAGS("00000018", ATOM_ELF_PH_FLAGS_STR)},
	{"alignment",			STATICTAG_EDIT_DWORD("0000001c")},
	{0, 0}
};

int_hash elf_ph_type[] =
{
	{ELF_PT_NULL, 		"null"},
	{ELF_PT_LOAD,		"load"},
	{ELF_PT_DYNAMIC,	"dynamic"},
	{ELF_PT_INTERP,	"interp"},
	{ELF_PT_NOTE,		"note"},
	{ELF_PT_SHLIB,		"shlib"},
	{ELF_PT_PHDR,		"phdr"},
	{0, 0}
};

ht_tag_flags_s elf_ph_flags[] =
{
	{0,  "[00] executable"},
	{1,  "[01] writable"},
	{2,  "[02] readable"},
	{0, 0}
};

ht_view *htelfprogramheaders_init(bounds *b, ht_stream *file, ht_format_group *group)
{
	ht_elf_shared_data *elf_shared=(ht_elf_shared_data *)group->get_shared_data();
	
	if ((elf_shared->header.e_ident[ELF_EI_CLASS]!=ELFCLASS32) || 
	(elf_shared->header.e_ident[ELF_EI_DATA]!=ELFDATA2LSB)) return 0;
	
	ht_uformat_view *v=new ht_uformat_view();
	v->init(b, DESC_ELF_PROGRAM_HEADERS, VC_EDIT, file, group);
	
	register_atom(ATOM_ELF_PH_TYPE, elf_ph_type);
	register_atom(ATOM_ELF_PH_FLAGS, elf_ph_flags);
	
	STREAMOFS h=elf_shared->header.e_phoff;
	
	hts_mask *m=new hts_mask();
	m->init(file, 0);
	
	char info[128];
	sprintf(info, "* ELF program headers at offset %08x", h);
	
	m->add_mask(info);

	v->insertsub(m);
	
	for (UINT i=0; i<elf_shared->pheaders.count; i++) {
		
		hts_mask *n=new hts_mask();
		n->init(file, i);
		
		char t[32];
		char *etype=matchhash(elf_shared->pheaders.pheaders[i].p_type, elf_ph_type);
		if (!etype) etype="?";
		sprintf(t, "entry %d (%s)", i, etype);
		
		n->add_staticmask_ptable(elfprogramheader, h+i*elf_shared->header.e_phentsize);
		
		hts_collapsable_sub *cn=new hts_collapsable_sub();
		cn->init(file, n, 1, t, 1);
	
		v->insertsub(cn);
	}
	
	return v;
}

format_view_if htelfprogramheaders_if = {
	htelfprogramheaders_init,
	0
};
