/*
 *	HT Editor
 *	htformat.h
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTFORMAT_H__
#define __HTFORMAT_H__

class ht_format_group;

#include "htobj.h"
#include "formats.h"

typedef dword fmt_vaddress;

/*
 *	commands
 */

#define cmd_view_mode		1000
#define cmd_edit_mode		1001
#define cmd_goto			1002
#define cmd_search			1003

/*
 *	CLASS ht_search_request
 */
 
class ht_search_request: public ht_data {
public:
	UINT search_class;
	UINT type;
	UINT flags;
	
			ht_search_request(UINT search_class, UINT type, UINT flags);
};

/*
 *	CLASS ht_search_result
 */
 
class ht_search_result: public ht_data {
public:
	UINT search_class;

			ht_search_result(UINT search_class);
};

/*
 *	CLASS ht_physical_search_result
 */
 
class ht_physical_search_result: public ht_search_result {
public:
	STREAMOFS offset;
	UINT size;
	
			ht_physical_search_result();
};

/*
 *	CLASS ht_visual_search_result
 */
 
class ht_visual_search_result: public ht_search_result {
public:
	fmt_vaddress address;
	UINT xpos;
	UINT length;
	
			ht_visual_search_result();
};

/*
 *	formats
 */

class ht_format_loc: public ht_data {
public:
	char *name;
	STREAMOFS start;
	UINT length;
};

/*
 *	vstate
 */

class vstate: public ht_data {
public:
	ht_view *focused;		/* view to focus */
	ht_view *view;			/* view whose state has changed */
	void *data;			/* view's data */
};

/*
 *	CLASS ht_viewer
 */

/* caps */
#define VC_NULL	0
#define VC_EDIT	1
#define VC_GOTO	2
#define VC_SEARCH	4
#define VC_BLOCKOP	8

class ht_viewer: public ht_view {
protected:
	UINT caps;

/* new */
	virtual	char *func(UINT i, bool execute);
public:
			void init(bounds *b, char *desc, UINT caps);
	virtual	void done();
/* overwritten */
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_format_view
 */

class ht_format_view: public ht_viewer {
protected:
	ht_stream *stream;
	ht_stack *vs_history;
	ht_format_group *format_group;
// last search (request)
	ht_search_request *last_search_request;
	bool last_search_physical;
	union {
		STREAMOFS	last_search_end_ofs;
		fmt_vaddress last_search_end_addr;
	};

/* new */
	virtual	bool address_to_offset(fmt_vaddress addr, STREAMOFS *ofs);
	virtual	bool offset_to_address(STREAMOFS ofs, fmt_vaddress *addr);
	/* view state histories */
			bool push_vs_history(ht_view *focused);
			bool pop_vs_history();
			
	virtual	void view_state_load(void *view_state);
	virtual	void *view_state_create();
	virtual	void view_state_destroy(void *view_state);
	
	virtual	bool next_logical_address(fmt_vaddress addr, fmt_vaddress *naddr);
	virtual	bool next_logical_offset(STREAMOFS ofs, STREAMOFS *nofs);
	virtual	bool show_search_result(ht_search_result *result);
public:
			void init(bounds *b, char *desc, UINT caps, ht_stream *stream, ht_format_group *format_group);
	virtual	void done();
/* overwritten */
	virtual	void handlemsg(htmsg *msg);
/* new */
	/* position indicator string */
	virtual	char *get_pindicator_str();
	/* physical file location(s) */
	virtual   bool loc_enum_init(ht_format_loc *start_below);
	virtual   bool loc_enum_next(ht_format_loc *loc);

	/* physical address (offset) functions */
	virtual	bool get_current_offset(STREAMOFS *ofs);
	virtual	bool goto_offset(STREAMOFS ofs, ht_view *source_object=0);
	virtual	UINT pread(STREAMOFS ofs, void *buf, UINT size);
	virtual	ht_search_result *psearch(ht_search_request *search, STREAMOFS start, STREAMOFS end);
	virtual	void pselect_add(STREAMOFS start, STREAMOFS end);
	virtual	void pselect_get(STREAMOFS *start, STREAMOFS *end);
	virtual	void pselect_set(STREAMOFS start, STREAMOFS end);
	virtual	UINT pwrite(STREAMOFS ofs, void *buf, UINT size);
	virtual	bool string_to_offset(char *string, STREAMOFS *ofs);

	/* virtual address functions */
	virtual	bool get_current_address(fmt_vaddress *addr);
	virtual	bool goto_address(fmt_vaddress addr, ht_view *source_object=0);
	virtual	UINT vread(fmt_vaddress addr, void *buf, UINT size);
	virtual	ht_search_result *vsearch(ht_search_request *search, fmt_vaddress start, fmt_vaddress end);
	virtual	void vselect_add(fmt_vaddress start, fmt_vaddress end);
	virtual	void vselect_get(fmt_vaddress *start, fmt_vaddress *end);
	virtual	void vselect_set(fmt_vaddress start, fmt_vaddress end);
	virtual	UINT vwrite(fmt_vaddress addr, void *buf, UINT size);
	virtual	bool string_to_address(char *string, fmt_vaddress *addr);
	
	/* search related */
			bool continue_search();
};

/*
 *	CLASS ht_format_group
 */

class ht_format_group: public ht_format_view {
protected:
	ht_stream *stream;
	format_view_if **ifs;
	ht_clist *format_views;
	void *shared_data;
	bool delete_stream;
	ht_xgroup *xgroup;

			void init_ifs(format_view_if **ifs);
			void done_ifs();
			
			bool init_if(format_view_if *i);
			bool done_if(format_view_if *i, ht_view *v);
public:
			void init(bounds *b, int options, char *desc, ht_stream *stream, bool delete_stream, format_view_if **ifs, ht_format_group *format_group);
	virtual		void done();
/* overwritten */
	virtual	int childcount();
	virtual	int focus(ht_view *view);
			void getbounds(bounds *b);
	virtual	void getclientarea(bounds *b);
	virtual 	ht_view *getfirstchild();
	virtual	ht_view *getselected();
	virtual	void handlemsg(htmsg *msg);
	virtual	void move(int x, int y);
	virtual	void receivefocus();
	virtual	void redraw();
	virtual	void releasefocus();
	virtual	void resize(int rw, int rh);
	virtual	void setgroup(ht_group *group);
/* new */
	virtual	void insert(ht_view *view);
			void remove(ht_view *view);
			void *get_shared_data();
};

/*
 *	CLASS ht_uformat_view
 */

#define cursor_state_visible		0
#define cursor_state_invisible	1
#define cursor_state_disabled		2

class hts_sub;

class ht_uformat_view: public ht_format_view {
protected:
public:
	bool edit;
	hts_sub *first_sub, *last_sub;
/* top line position */
	hts_sub *top_sub;
	ID top_id1, top_id2;
/* cursor line and state and tag position */
	int cursor_ypos;
	int cursor_state;
	bool cursor_select;
	STREAMOFS cursor_select_start;
	dword cursor_select_cursor_length;
	hts_sub *cursor_sub;
	ID cursor_id1, cursor_id2;
	int cursor_tag_idx;
	int cursor_tag_group;
	int cursor_tag_micropos;
/* selection*/
	STREAMOFS sel_start;
	STREAMOFS sel_end;
/* visual info */
	int cursor_visual_xpos;
	int cursor_visual_length;
/* misc info */
	int cursor_tag_class;
	union {
		STREAMOFS cursor_tag_offset;
		struct {
			ID low, high;
		} cursor_tag_id;
	};
	char cursor_line[1024];	/* FIXME: possible buffer overflow ! */
/* tag palettes */
	palette tagpal;
/**/
	int xscroll;

	bool uf_initialized;

	bool isdirty_cursor_line;

	char pind_str[64];

/* overwritten */
	virtual	char *func(UINT i, bool execute);
	virtual	bool next_logical_address(fmt_vaddress addr, fmt_vaddress *naddr);
	virtual	bool next_logical_offset(STREAMOFS ofs, STREAMOFS *nofs);
	virtual	void view_state_load(void *view_state);
	virtual	void *view_state_create();
/* new */
	void adjust_cursor_group();
	void adjust_cursor_idx();
	int center_view(hts_sub *sub, ID id1, ID id2);
	void check_cursor_visibility();
	void complete_init();
	int cursor_left();
	int cursor_right();
	int cursor_down(int n);
	int cursor_up(int n);
	int cursor_home();
	int cursor_end();
	void cursor_tab();
	void cursorline_dirty();
	void cursorline_get();
	virtual int cursormicro_forward();
	virtual int cursormicro_backward();
	virtual int cursormicroedit_forward();
	bool edit_end();
	void edit_input(byte b);
	int edit_input_c2h(byte b);
	void edit_input_correctpos();
	bool edit_start();
	bool edit_update();
	int find_first_tag(hts_sub **sub, ID *id1, ID *id2, int limit);
	int find_first_edit_tag_with_offset(hts_sub **sub, ID *id1, ID *id2, int *tag_idx, int limit, STREAMOFS offset);
	void focus_cursor();
	vcp getcolor_tag(UINT pal_index);
	int get_current_tag(char **tag);
	int get_current_tag_size(dword *size);
	vcp get_tag_color_edit(STREAMOFS tag_offset, UINT size, bool iscursor);
	int next_line(hts_sub **sub, ID *id1, ID *id2, int n);
	int prev_line(hts_sub **sub, ID *id1, ID *id2, int n);
	void print_tagstring(int x, int y, int maxlen, int xscroll, char *tagstring, bool cursor_in_line);
	int ref();
	int ref_desc(dword id, STREAMOFS offset, int size);
	int ref_flags(dword id, STREAMOFS offset);
	virtual int ref_sel(ID id_low, dword id_high);
	UINT render_tagstring(char *chars, vcp *colors, UINT maxlen, char *tagstring, bool cursor_in_line);
	void render_tagstring_desc(char **string, int *length, vcp *tag_color, char *tag, UINT size, bool is_cursor);
	UINT render_tagstring_single(char *chars, vcp *colors, UINT maxlen, UINT offset, char *text, UINT len, vcp color);
	void scroll_up(int n);
	void scroll_down(int n);
	void select_mode_off();
	void select_mode_on();
	void select_mode_pre();
	void select_mode_post(bool lastpos);
	int set_cursor(hts_sub *sub, ID id1, ID id2);
	int set_cursor_ex(hts_sub *sub, ID id1, ID id2, int tag_idx, int tag_group);
	void update_micropos();
	void update_misc_info();
	void update_visual_info();
	void update_ypos();
public:
	UINT search_caps;
	
			void init(bounds *b, char *desc, int caps, ht_stream *stream, ht_format_group *format_group);
	virtual	void done();
/* overwritten */
	virtual	void draw();
	virtual	bool get_current_address(fmt_vaddress *address);
	virtual	bool get_current_offset(STREAMOFS *offset);
	virtual	char *get_pindicator_str();
	virtual	bool goto_offset(STREAMOFS offset, ht_view *source_object=0);
	virtual	bool goto_address(fmt_vaddress addr, ht_view *source_object=0);
	virtual	void handlemsg(htmsg *msg);
	virtual	ht_search_result *psearch(ht_search_request *search, STREAMOFS start, STREAMOFS end);
	virtual	void pselect_add(STREAMOFS start, STREAMOFS end);
	virtual	void pselect_get(STREAMOFS *start, STREAMOFS *end);
	virtual	void pselect_set(STREAMOFS start, STREAMOFS end);
	virtual	bool string_to_address(char *string, fmt_vaddress *vaddr);
	virtual	ht_search_result *vsearch(ht_search_request *search, fmt_vaddress start, fmt_vaddress end);
/* new */
	virtual	void insertsub(hts_sub *sub);
};

/*
 *	CLASS hts_sub
 */

/* search results */
#define SR_NOT_FOUND		0
#define SR_FOUND			1
#define SR_NOT_SUPPORTED		2

class hts_sub: public object {
protected:
	ht_stream *stream;
public:
	hts_sub *prev, *next;

			void init(ht_stream *stream);
	virtual	void done();
/* new */
	virtual	bool convert_addr_to_id(fmt_vaddress addr, ID *id1, ID *id2);
	virtual	bool convert_ofs_to_id(STREAMOFS offset, ID *id1, ID *id2);
	virtual	bool convert_id_to_addr(ID id1, ID id2, fmt_vaddress *addr);
	virtual	bool convert_id_to_ofs(ID id1, ID id2, STREAMOFS *offset);
	virtual	void first_line_id(ID *id1, ID *id2);
	virtual	bool getline(char *line, ID id1, ID id2);
	virtual	void last_line_id(ID *id1, ID *id2);
	virtual	int next_line_id(ID *id1, ID *id2, int n);
	virtual	int prev_line_id(ID *id1, ID *id2, int n);
	virtual	bool ref(ID id1, ID id2);
	virtual	ht_search_result *search(ht_search_request *search, STREAMOFS start, STREAMOFS end);
};

/*
 *	CLASS hts_linear_sub
 */

class hts_linear_sub: public hts_sub {
protected:
	STREAMOFS fofs;
	dword fsize;
public:
			void init(ht_stream *stream, STREAMOFS offset, int size);
	virtual	void done();
/* overwritten */
	virtual	ht_search_result *search(ht_search_request *search, STREAMOFS start, STREAMOFS end);
};

/*
 *	CLASS hts_hex
 */

class hts_hex: public hts_linear_sub {
protected:
	dword vaddrinc;
	dword balign;
public:
			void init(ht_stream *stream, STREAMOFS ofs, dword size, dword vaddrinc=0);
	virtual 	void done();
/* overwritten */
	virtual	bool convert_addr_to_id(fmt_vaddress addr, ID *id1, ID *id2);
	virtual	bool convert_ofs_to_id(STREAMOFS offset, ID *id1, ID *id2);
	virtual	bool convert_id_to_addr(ID id1, ID id2, fmt_vaddress *addr);
	virtual bool convert_id_to_ofs(ID id1, ID id2, STREAMOFS *ofs);
	virtual	void first_line_id(ID *id1, ID *id2);
	virtual	bool getline(char *line, ID id1, ID id2);
	virtual	void last_line_id(ID *id1, ID *id2);
	virtual	int next_line_id(ID *id1, ID *id2, int n);
	virtual	int prev_line_id(ID *id1, ID *id2, int n);
};

/*
 *	CLASS hts_mask
 */

struct hts_mask_ptable {
	char *desc;
	char *fields;
};
 
class hts_mask: public hts_sub {
protected:
	ht_string_list *masks;
	char temp[512];
	UINT uid;
public:
			void init(ht_stream *stream, UINT uid);
	virtual 	void done();
/* overwritten */
	virtual 	bool getline(char *line, ID id1, ID id2);
	virtual 	void first_line_id(ID *id1, ID *id2);
	virtual 	void last_line_id(ID *id1, ID *id2);
	virtual 	int next_line_id(ID *id1, ID *id2, int n);
	virtual 	int prev_line_id(ID *id1, ID *id2, int n);
/* new */
	virtual 	void add_mask(char *tagstr);
	virtual 	void add_mask_table(char **tagstr);
	virtual 	void add_staticmask(char *statictag_str, STREAMOFS reloc);
	virtual 	void add_staticmask_table(char **statictag_table, STREAMOFS reloc);
	virtual 	void add_staticmask_ptable(hts_mask_ptable *statictag_ptable, STREAMOFS reloc);
};

/*
 *	CLASS hts_layer_sub
 */

class hts_layer_sub: public hts_sub {
protected:
	hts_sub *sub;
	bool own_sub;
public:
			void init(ht_stream *stream, hts_sub *sub, bool own_sub);
	virtual	void done();
/* new */
	virtual	bool convert_addr_to_id(fmt_vaddress addr, ID *id1, ID *id2);
	virtual	bool convert_ofs_to_id(STREAMOFS offset, ID *id1, ID *id2);
	virtual	bool convert_id_to_addr(ID id1, ID id2, fmt_vaddress *addr);
	virtual	bool convert_id_to_ofs(ID id1, ID id2, STREAMOFS *offset);
	virtual	void first_line_id(ID *id1, ID *id2);
	virtual	bool getline(char *line, ID id1, ID id2);
	virtual	void last_line_id(ID *id1, ID *id2);
	virtual	int next_line_id(ID *id1, ID *id2, int n);
	virtual	int prev_line_id(ID *id1, ID *id2, int n);
	virtual	bool ref(ID id1, ID id2);
	virtual	ht_search_result *search(ht_search_request *search, STREAMOFS start, STREAMOFS end);
};

/*
 *	CLASS hts_collapsable_sub
 */

class hts_collapsable_sub: public hts_layer_sub {
protected:
	char *nodestring;
	bool collapsed;
	ID fid1, fid2;
	ID myfid1, myfid2;
	fmt_vaddress myfaddr;
public:
			void init(ht_stream *stream, hts_sub *sub, bool own_sub, char *nodename, bool collapsed);
	virtual	void done();
/* new */
	virtual	bool convert_addr_to_id(fmt_vaddress addr, ID *id1, ID *id2);
	virtual	bool convert_ofs_to_id(STREAMOFS offset, ID *id1, ID *id2);
	virtual	bool convert_id_to_addr(ID id1, ID id2, fmt_vaddress *addr);
	virtual	bool convert_id_to_ofs(ID id1, ID id2, STREAMOFS *offset);
	virtual	void first_line_id(ID *id1, ID *id2);
	virtual	bool getline(char *line, ID id1, ID id2);
	virtual	void last_line_id(ID *id1, ID *id2);
	virtual	int next_line_id(ID *id1, ID *id2, int n);
	virtual	int prev_line_id(ID *id1, ID *id2, int n);
	virtual	bool ref(ID id1, ID id2);
	virtual	ht_search_result *search(ht_search_request *search, STREAMOFS start, STREAMOFS end);
};

/*
 *	CLASS hts_unasm
 */

#include "asm.h"

class hts_unasm: public hts_linear_sub {
private:
	disassembler *unasm;
	int display_style;
public:
			void init(ht_stream *file, STREAMOFS ofs, int size, disassembler *dis, int display_style);
	virtual 	void done();
/* overwritten */
	virtual	bool convert_addr_to_id(fmt_vaddress addr, ID *id1, ID *id2);
	virtual	bool convert_ofs_to_id(STREAMOFS offset, ID *id1, ID *id2);
	virtual	bool convert_id_to_addr(ID id1, ID id2, fmt_vaddress *addr);
	virtual	void first_line_id(ID *id1, ID *id2);
	virtual	bool getline(char *line, ID id1, ID id2);
	virtual	void last_line_id(ID *id1, ID *id2);
	virtual	int next_line_id(ID *id1, ID *id2, int n);
	virtual	int prev_line_id(ID *id1, ID *id2, int n);
};

/*
 *	CLASS ht_data_tagstring
 */

class ht_data_tagstring: public ht_data_string {
public:
			ht_data_tagstring(char *tagstr=0);
	virtual	~ht_data_tagstring();
};

#endif /* __HTFORMAT_H__ */
