/* 
 *	HT Editor
 *	hthist.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htatom.h"
#include "htdata.h"
#include "hthist.h"

bool insert_history_entry(char *name, void *data, ht_list *history)
{
	ht_history_entry *e=new ht_history_entry(ht_strdup(name), data);
	UINT li=history->find(e);
	if (li==LIST_UNDEFINED) {
		history->prepend(e);
		return 1;
	} else {
		delete e;
		history->move(li, 0);
		return 0;
	}							
}

/*
 *	CLASS ht_history_entry
 */

ht_history_entry::ht_history_entry(char *str, void *_data)
{
	desc=new ht_data_string(str);
	data=_data;
}

ht_history_entry::~ht_history_entry()
{
	delete desc;
}

int compare_keys_history_entry(ht_data *key_a, ht_data *key_b)
{
    return compare_keys_string(((ht_history_entry*)key_a)->desc, ((ht_history_entry*)key_b)->desc);
}

/*
 *	INIT
 */

void create_hist_atom(UINT atom)
{
	ht_clist *c=new ht_clist();
	c->init(compare_keys_history_entry);
	register_atom(atom, c);
}

void destroy_hist_atom(UINT atom)
{
	ht_clist *c=(ht_clist*)find_atom(atom);
	unregister_atom(atom);
	c->done();
	delete c;
}

void init_hist()
{
	create_hist_atom(HISTATOM_GOTO);
	create_hist_atom(HISTATOM_FILE);
	create_hist_atom(HISTATOM_SEARCH_BIN);
	create_hist_atom(HISTATOM_SEARCH_CASM);
	create_hist_atom(HISTATOM_SEARCH_TEXT);
	create_hist_atom(HISTATOM_ASSEMBLER);
}

/*
 *	DONE
 */

void done_hist()
{
	destroy_hist_atom(HISTATOM_ASSEMBLER);
	destroy_hist_atom(HISTATOM_SEARCH_TEXT);
	destroy_hist_atom(HISTATOM_SEARCH_CASM);
	destroy_hist_atom(HISTATOM_SEARCH_BIN);
	destroy_hist_atom(HISTATOM_FILE);
	destroy_hist_atom(HISTATOM_GOTO);
}

