/*
 *	HT Editor
 *	htreg.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htreg.h"

#include <string.h>

ht_registry *registry;

/*
 *	CLASS ht_registry_node
 */

void ht_registry_node::init(ht_registry_node_type _type)
{
	type=_type;
	data=0;
}

void ht_registry_node::done()
{
	if (data) {
		data->done();
		delete data;
	}
}

/*
 *	CLASS ht_registry
 */

void ht_registry::init()
{
// build registry root
	root=new ht_registry_node();

	root->init(RNT_SUBDIR);

	ht_stree *s=new ht_stree();
	s->init(compare_keys_string);

	root->data=s;

// build node_types tree
	node_types=new ht_stree();
	node_types->init(compare_keys_string);

	node_types->insert(new ht_data_string("subdir"), new ht_data_dword(RNT_SUBDIR));
	node_types->insert(new ht_data_string("symlink"), new ht_data_dword(RNT_SYMLINK));
	node_types->insert(new ht_data_string("dword"), new ht_data_dword(RNT_DWORD));
	node_types->insert(new ht_data_string("string"), new ht_data_dword(RNT_STRING));
	node_types->insert(new ht_data_string("mem"), new ht_data_dword(RNT_MEM));
}

void ht_registry::done()
{
	node_types->done();
	delete node_types;

	root->done();
	delete root;
}

bool ht_registry::create_dword(char *key)
{
	return create_node(key, RNT_DWORD);
}

bool ht_registry::create_mem(char *key)
{
	return create_node(key, RNT_MEM);
}

bool ht_registry::create_node(char *key, UINT type)
{
	char *name;
	ht_registry_node *m;
	if (!splitfind(key, &name, &m)) return 0;

	ht_registry_node *n=new ht_registry_node();
	n->init(type);

	((ht_stree*)m->data)->insert(new ht_data_string(name), n);
	return 1;
}

bool ht_registry::create_string(char *key)
{
	return create_node(key, RNT_STRING);
}

bool ht_registry::create_subdir(char *key)
{
	char *name;
	ht_registry_node *m;
	if (!splitfind(key, &name, &m)) return 0;

	ht_stree *s=new ht_stree();
	s->init(compare_keys_string);

	ht_registry_node *n=new ht_registry_node();
	n->init(RNT_SUBDIR);
	n->data=s;

	((ht_stree*)m->data)->insert(new ht_data_string(name), n);
	return 1;
}

bool ht_registry::create_symlink(char *key)
{
	return create_node(key, RNT_SYMLINK);
}

void ht_registry::debug_dump()
{
//	FILE *f=fopen("", "");
	debug_dump_i(stderr, (ht_tree*)root->data, 0);
//	fclose(f);
}

void ht_registry::debug_dump_i(FILE *f, ht_tree *t, int ident)
{
	ht_data_string *key;
	ht_registry_node *n;
	if (!t->enum_first((ht_data**)&key, (ht_data**)&n)) return;
	do {
		for (int i=0; i<ident; i++) fprintf(f, "     ");
		fprintf(f, "%s ", key->value);
		switch (n->type) {
			case RNT_DWORD:
				fprintf(f, "= (dword) %08d (%08x)\n", ((ht_data_dword*)n->data)->value, ((ht_data_dword*)n->data)->value);
				break;
			case RNT_STRING:
				fprintf(f, "= (string) \"%s\"\n", ((ht_data_string*)n->data)->value);
				break;
			case RNT_SYMLINK:
				fprintf(f, "=> \"%s\"\n", ((ht_data_string*)n->data)->value);
				break;
			case RNT_SUBDIR:
				fprintf(f, "{\n");
				debug_dump_i(f, (ht_tree*)n->data, ident+1);
				for (int i=0; i<ident; i++) fprintf(f, "     ");
				fprintf(f, "}\n");
				break;
			case RNT_MEM:
				fprintf(f, "= (mem) nyi!\n");
				break;
			default: {
				char *name=lookup_node_type_name(n->type);
				if (!name) name="?";
				fprintf(f, "= ('%s'=%d)\n", name, n->type);
				break;
			}
		}
	} while (t->enum_next((ht_data**)&key, (ht_data**)&n));
}

bool ht_registry::enum_first(char *key, char **keyname, ht_data **data, ht_registry_node_type *type)
{
	rec_depth=0;
	ht_registry_node *n=find_entry_i(key, 1);
	if (n) {
		if (n->type!=RNT_SUBDIR) return 0;
		enum_tree=(ht_stree*)n->data;
		ht_data_string *k;
		ht_registry_node *d;
		if (!enum_tree->enum_first((ht_data**)&k, (ht_data**)&d)) return 0;
		*keyname=k->value;
		*data=d->data;
		*type=d->type;
		return 1;
	}
	return 0;
}

bool ht_registry::enum_next(char **keyname, ht_data **data, ht_registry_node_type *type)
{
	ht_data_string *k;
	ht_registry_node *d;
	if (enum_tree->enum_next((ht_data**)&k, (ht_data**)&d)) {
		*keyname=k->value;
		*data=d->data;
		*type=d->type;
		return 1;
	}
	return 0;
}

bool ht_registry::find_any_entry(char *key, ht_data **data, ht_registry_node_type *type)
{
	rec_depth=0;
	ht_registry_node *n=find_entry_i(key, 1);
	if (n) {
		*data=n->data;
		*type=n->type;
		return 1;
	}
	return 0;
}

bool ht_registry::find_data_entry(char *key, ht_data **data, ht_registry_node_type *type)
{
	rec_depth=0;
	ht_registry_node *n=find_entry_i(key, 1);
	if (n) {
		if (n->type==RNT_SUBDIR) return 0;
		*data=n->data;
		*type=n->type;
		return 1;
	}
	return 0;
}

ht_registry_node *ht_registry::find_entry_i(char *key, int follow_symlinks)
{
	ht_registry_node *dir=root;
	char *s;
	char t[256]; /* FIXME: possible buffer overflow */
	if (key[0]=='/') key++;
	while (1) {
		s=strchr(key, '/');
		if (s) {
			strncpy(t, key, s-key);
			t[s-key]=0;
			dir=find_entry_get_subdir((ht_tree*)dir->data, t);
			if (!dir) break;
			key=s+1;
		} else {
			ht_registry_node *n;
			if (*key==0) {
				n=dir;
			} else {
				n=find_entry_get_data((ht_tree*)dir->data, key, follow_symlinks);
			}
			return n;
		}
	}
	return 0;
}

ht_registry_node *ht_registry::find_entry_get_node(ht_tree *dir, char *nodename)
{
	ht_data_string keystr(nodename);
	ht_registry_node *n=(ht_registry_node*)dir->get(&keystr);
	return n;
}

ht_registry_node *ht_registry::find_entry_get_subdir(ht_tree *dir, char *nodename)
{
	ht_registry_node *n=find_entry_get_node(dir, nodename);
start:
	if (!n) return 0;
	switch (n->type) {
		case RNT_SYMLINK: {
			rec_depth++;
			if (rec_depth>MAX_REC_DEPTH) return 0;
			char *sl=((ht_data_string*)n->data)->value;
			if (sl[0]=='/') {
				n=find_entry_i(sl, 1);
				goto start;
			} else {
				return find_entry_get_subdir(dir, sl);
			}
		}
		case RNT_SUBDIR:
			return n;
	}
	return 0;
}

ht_registry_node *ht_registry::find_entry_get_data(ht_tree *dir, char *nodename, int follow_symlinks)
{
	ht_registry_node *n=find_entry_get_node(dir, nodename);
start:
	if (!n) return 0;
	if ((n->type==RNT_SYMLINK) && (follow_symlinks)) {
		rec_depth++;
		if (rec_depth>MAX_REC_DEPTH) return 0;
		char *sl=((ht_data_string*)n->data)->value;
		if (sl[0]=='/') {
			n=find_entry_i(sl, 1);
			goto start;
		} else {
			return find_entry_get_data(dir, sl, follow_symlinks);
		}
	}
	return n;
}

ht_registry_node_type ht_registry::have_node_type(char *identifier)
{
	ht_registry_node_type t=lookup_node_type(identifier);
	if (!t) t=register_node_type(identifier);
	return t;
}

ht_registry_node_type ht_registry::lookup_node_type(char *identifier)
{
	ht_data_string s(identifier);
	ht_data_dword *d=(ht_data_dword*)node_types->get(&s);
	if (d) return (ht_registry_node_type)d->value;
	return RNT_INVALID;
}

char *ht_registry::lookup_node_type_name(ht_registry_node_type t)
{
	ht_data_string *key;
	ht_data_dword *data;
	if (node_types->enum_first((ht_data**)&key, (ht_data**)&data)) {
		do {
			if (t==data->value) return key->value;
		} while (node_types->enum_next((ht_data**)&key, (ht_data**)&data));
	}
	return 0;
}

ht_registry_node_type ht_registry::register_node_type(char *identifier)
{
	ht_registry_node_type t=0;
	ht_data_string *key;
	ht_data_dword *data;
	bool found=0;
	while (!found) {
		t++;
		found=1;
		if (node_types->enum_first((ht_data**)&key, (ht_data**)&data)) {
			do {
				if (t==data->value) {
					found=0;
					break;
				}
			} while (node_types->enum_next((ht_data**)&key, (ht_data**)&data));
		}
	}
	bool b=node_types->insert(new ht_data_string(identifier), new ht_data_dword(t));
	if (b) return t;
	return RNT_INVALID;
}

bool ht_registry::set_dword(char *key, dword d)
{
	return set_node(key, RNT_DWORD, new ht_data_dword(d));
}

bool ht_registry::set_mem(char *key, void *mem, UINT size)
{
	return set_node(key, RNT_MEM, new ht_data_mem(mem, size));
}

bool ht_registry::set_node(char *key, ht_registry_node_type type, ht_data *data)
{
	ht_registry_node *n=find_entry_i(key, 0);
	if (!n) return 0;

	if (n->type==type) {
		if (n->data) {
			n->data->done();
			delete n->data;
		}
		n->data=data;
		return 1;
	}
	return 0;
}

bool ht_registry::set_string(char *key, char *string)
{
	return set_node(key, RNT_STRING, new ht_data_string(string));
}

bool ht_registry::set_symlink(char *key, char *dest)
{
	return set_node(key, RNT_SYMLINK, new ht_data_string(dest));
}

bool ht_registry::splitfind(char *key, char **name, ht_registry_node **node)
{
	char dir[256]; /* FIXME: possible buffer overflow */
	char *n=strrchr(key, '/');
	if (n) {
		strncpy(dir, key, n-key);
		dir[n-key]=0;
		n++;
	} else {
		dir[0]=0;
		n=key;
	}

	ht_registry_node *m=find_entry_i(dir, 1);
	if (!m) return 0;
	if (m->type!=RNT_SUBDIR) return 0;
	*node=m;
	*name=n;
	return 1;
}

/*
 *	INIT
 */

int init_registry()
{
	registry=new ht_registry();
	registry->init();
	return 1;
}

/*
 *	DONE
 */

int done_registry()
{
	registry->done();
	delete registry;
	return 1;
}

