/* 
 *	HT Editor
 *	htreg.h
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTREG_H__
#define __HTREG_H__

#include "htdata.h"
#include "htstring.h"

/*
 *	CLASS ht_registry_node
 */

typedef UINT ht_registry_node_type;

#define RNT_INVALID		0	/* returned by some functions */
// these are predefined
#define RNT_SUBDIR		1
#define RNT_SYMLINK		2
#define RNT_DWORD  		3
#define RNT_STRING 		4
#define RNT_MEM		5
// the rest may be allocated dynamically

class ht_registry_node: public ht_data {
public:
	ht_registry_node_type type;
	ht_data *data;

			void init(ht_registry_node_type type);
	virtual	void done();
};

/*
 *	CLASS ht_registry
 */

#define MAX_REC_DEPTH 20

class ht_registry: public ht_data {
protected:
	ht_registry_node *root;
	ht_stree *node_types;
	ht_stree *enum_tree;
	UINT rec_depth;

			bool find_any_entry(char *key, ht_data **data, ht_registry_node_type *type);

			ht_registry_node *find_entry_i(char *key, int follow_symlinks);
			ht_registry_node *find_entry_get_node(ht_tree *dir, char *nodename);
			ht_registry_node *find_entry_get_subdir(ht_tree *dir, char *nodename);
			ht_registry_node *find_entry_get_data(ht_tree *dir, char *nodename, int follow_symlinks);
			bool splitfind(char *key, char **name, ht_registry_node **node);
public:

			void init();
	virtual	void done();
/* new */
			bool create_dword(char *key);
			bool create_mem(char *key);
			bool create_node(char *key, ht_registry_node_type type);
			bool create_string(char *key);
			bool create_subdir(char *key);
			bool create_symlink(char *key);
			bool enum_first(char *key, char **keyname, ht_data **data, ht_registry_node_type *type);
			bool enum_next(char **keyname, ht_data **data, ht_registry_node_type *type);
			bool find_data_entry(char *key, ht_data **data, ht_registry_node_type *type);
			ht_registry_node_type lookup_node_type(char *identifier);
			char *lookup_node_type_name(ht_registry_node_type t);
			ht_registry_node_type have_node_type(char *identifier);
			ht_registry_node_type register_node_type(char *identifier);
			bool set_dword(char *key, dword d);
			bool set_mem(char *key, void *mem, UINT size);
			bool set_node(char *key, ht_registry_node_type type, ht_data *data);
			bool set_string(char *key, char *string);
			bool set_symlink(char *key, char *dest);
/* debug */
			void debug_dump();
			void debug_dump_i(FILE *f, ht_tree *t, int ident);
};

extern ht_registry *registry;

/*
 *	INIT
 */

int init_registry();

/*
 *	DONE
 */

int done_registry();

#endif /* __HTREG_H__ */
