/* 
 *	HT Editor
 *	htsearch.h
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTSEARCH_H__
#define __HTSEARCH_H__

#include "htdialog.h"
#include "htobj.h"
#include "htstream.h"

#include <sys/types.h>
extern "C" {
#include "regex.h"
}

/*
 *	searching
 */

// search classes
#define SC_PHYSICAL			0    // search in underlying binary data
#define SC_VISUAL			1    // search in displayed text

// search types + flags
#define ST_FXBIN		0		// search using fixed binary pattern
	#define SF_FXBIN_CASEINSENSITIVE		1
	#define SF_FXBIN_IDLE				2
	
#define ST_REGEX		1         // search using regular expression
	#define SF_REGEX_CASEINSENSITIVE		1
	#define SF_REGEX_IDLE				2

/*
 *	CLASS ht_fxbin_search_request
 */
 
class ht_fxbin_search_request: public ht_search_request {
public:
	UINT data_size;
	byte *data;
	
			ht_fxbin_search_request(UINT search_class, UINT flags, UINT data_size, byte *data);
	virtual	~ht_fxbin_search_request();
/* overwritten */
	virtual object *duplicate();
};

/*
 *	CLASS ht_regex_search_request
 */
 
class ht_regex_search_exception {
public:
	int errorcode;
	regex_t *regex;
	
	ht_regex_search_exception(int e, regex_t *r) { errorcode=e; regex=r; }
};

class ht_regex_search_request: public ht_search_request {
public:
	char *rx_str;
	regex_t rx;

			ht_regex_search_request(UINT search_class, UINT flags, char *regex);
	virtual	~ht_regex_search_request();
/* overwritten */
	virtual object *duplicate();
};

/* binary search function */

// flags
#define SFBIN_CASEINSENSITIVE	1
#define SFBIN_IDLE			2

int search_bin(ht_stream *stream, STREAMOFS start, STREAMOFS end, dword pattern_len, byte *pattern, STREAMOFS *result_ofs, UINT flags, ht_progress_indicator *progress_indicator);

/*
 *	CLASS ht_hexascii_search_mask
 */

struct ht_hexascii_search_mask_data {
	byte *str;
	dword strlen;
	byte *start;
	dword startlen;
	byte *end;
	dword endlen;
	int caseinsensitive;
	int idle;
	char variable[0];
};

class ht_hexascii_search_mask: public ht_group {
protected:
	ht_strinputfield *str;
	ht_strinputfield *range_start;
	ht_strinputfield *range_end;
	ht_checkboxes *option_boxes;
public:
			void init(bounds *b, int options, ht_list *history=0);
	virtual	void done();
/* overwritten */
	virtual int datasize();
	virtual void getdata(void *buf);
	virtual void setdata(void *buf);
};

/*
 *	CLASS ht_casm_search_mask
 */

struct ht_casm_search_mask_data {
	byte *str;
	dword strlen;
	byte *start;
	dword startlen;
	byte *end;
	dword endlen;
	int caseinsensitive;
	int idle;
	char variable[0];
};

class ht_casm_search_mask: public ht_group {
protected:
	ht_strinputfield *str;
	ht_strinputfield *range_start;
	ht_strinputfield *range_end;
	ht_checkboxes *option_boxes;
public:
			void	init(bounds *b, int options, ht_list *history=0);
	virtual	void done();
/* overwritten */
	virtual int datasize();
	virtual void getdata(void *buf);
	virtual void setdata(void *buf);
};

/*
 *	CLASS ht_text_search_mask
 */

struct ht_text_search_mask_data {
	byte *str;
	dword strlen;
	byte *start;
	dword startlen;
	byte *end;
	dword endlen;
	int caseinsensitive;
	int idle;
	char variable[0];
};

class ht_text_search_mask: public ht_group {
protected:
	ht_strinputfield *str;
	ht_strinputfield *range_start;
	ht_strinputfield *range_end;
	ht_checkboxes *option_boxes;
public:
			void	init(bounds *b, int options, ht_list *history=0);
	virtual	void done();
/* overwritten */
	virtual int datasize();
	virtual void getdata(void *buf);
	virtual void setdata(void *buf);
};

#endif /* __HTSEARCH_H__ */
