/* 
 *	HT Editor
 *	htstream.h
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTSTREAM_H__
#define __HTSTREAM_H__

#include "global.h"
#include "htdata.h"

#include <stdio.h>
#include <sys/types.h>
#include <time.h>

typedef unsigned int STREAMOFS;

/*
 *	STRUCT ht_pstat	(portable stat)
 */

#if defined(WIN32) || defined(__WIN32__)
#define HT_S_IRUSR		S_IRUSR
#define HT_S_IRGRP		0
#define HT_S_IROTH      0
#define HT_S_IWUSR      S_IWUSR
#define HT_S_IWGRP      0
#define HT_S_IWOTH      0
#define HT_S_IXUSR		S_IXUSR
#define HT_S_IXGRP      0
#define HT_S_IXOTH		0
#define HT_S_IRWXU      S_IRWXU
#define HT_S_IRWXG		0
#define HT_S_IRWXO		0
#else
#define HT_S_IRUSR		S_IRUSR
#define HT_S_IRGRP		S_IRGRP
#define HT_S_IROTH      S_IROTH
#define HT_S_IWUSR      S_IWUSR
#define HT_S_IWGRP      S_IWGRP
#define HT_S_IWOTH      S_IWOTH
#define HT_S_IXUSR		S_IXUSR
#define HT_S_IXGRP		S_IXGRP
#define HT_S_IXOTH		S_IXOTH
#define HT_S_IRWXU      S_IRWXU
#define HT_S_IRWXG		S_IRWXG
#define HT_S_IRWXO		S_IRWXO
#endif

#define pstat_ctime			0x00000001
#define pstat_mtime			0x00000002
#define pstat_atime			0x00000004
#define pstat_uid			0x00000008
#define pstat_gid			0x00000010
#define pstat_mode_usr		0x00000020
#define pstat_mode_grp		0x00000040
#define pstat_mode_oth		0x00000080
#define pstat_mode_r          0x00000100
#define pstat_mode_w          0x00000200
#define pstat_mode_x          0x00000400
#define pstat_size			0x00000800
#define pstat_inode			0x00001000
#define pstat_cluster		0x00002000
#define pstat_fsid			0x00004000

#define pstat_mode_all		(pstat_mode_usr|pstat_mode_grp|pstat_mode_oth|pstat_mode_r|pstat_mode_w|pstat_mode_x)

struct ht_pstat {
	dword	caps;
	time_t	ctime;
	time_t	mtime;
	time_t	atime;
	int		uid;
	int		gid;
	mode_t	mode;	// S_ISUID, S_ISGID, S_I[RWX](USR|GRP|OTH)
	dword	size;
	dword	fsid;	// inode OR cluster OR fsid
};

/*
 *	CLASS ht_stream
 */

#define SERR_FAIL			0
#define SERR_RETRY			1

#define STERR_SYSTEM		0x80000000
#define STERR_INTERNAL		0x40000000

/* ht_stream.access_mode */
#define SAM_UNDEFINED		0
#define SAM_READ			1
#define SAM_WRITE			2
#define SAM_CREATE			4

/* ht_stream.mode */
#define SM_READ_CACHED		1

#define ISDIRTY_SINGLEBIT	0x80000000

class ht_stream {
protected:
	int (*stream_error_proc)(ht_stream *stream);
	int error;
	dword access_mode;
	dword mode;

	int call_error_proc();
	void set_error(int error);
public:
			void	init();
	virtual	void done();
/* new */
	virtual void flush();
	virtual int get_error();
	virtual dword get_mode();
	virtual char *get_name();
	virtual dword get_size();
	virtual void invalidate();
	virtual int isdirty(STREAMOFS offset, int range);
	virtual void pstat(ht_pstat *s);
	virtual int seek(STREAMOFS offset);
	virtual int set_access_mode(dword access_mode);
	virtual void set_error_proc(int (*stream_error_proc)(ht_stream *stream));
	virtual void set_mode(dword mode);
	virtual STREAMOFS tell();

	virtual dword read(void *buf, dword size);
	virtual dword write(void *buf, dword size);
};

/*
 *	ht_layer_stream
 */

class ht_layer_stream: public ht_stream {
protected:
	ht_stream *stream;
	int own_stream;
public:
			void	init(ht_stream *stream, int own_stream);
	virtual	void done();
/* overwritten */
	virtual void flush();
	virtual int get_error();
	virtual dword get_mode();
	virtual char *get_name();
	virtual dword get_size();
	virtual void invalidate();
	virtual int isdirty(STREAMOFS offset, int range);
	virtual void pstat(ht_pstat *s);
	virtual int seek(STREAMOFS offset);
	virtual void set_mode(dword mode);
	virtual int set_access_mode(dword access_mode);
	virtual void set_error_proc(int (*stream_error_proc)(ht_stream *stream));
	virtual STREAMOFS tell();

	virtual dword read(void *buf, dword size);
	virtual dword write(void *buf, dword size);
};

/*
 *	CLASS ht_file_stream
 */

class ht_file_stream: public ht_stream {
protected:
	char *name;
	STREAMOFS offset;
	FILE *file;

	int set_access_mode_internal(dword access_mode);
public:

			void	init(const char *name, int access_mode);
	virtual	void done();
/* overwritten */
	virtual char *get_name();
	virtual dword get_size();
	virtual void pstat(ht_pstat *s);
	virtual int seek(STREAMOFS offset);
	virtual int set_access_mode(dword access_mode);
	virtual STREAMOFS tell();

	virtual dword read(void *buf, dword size);
	virtual dword write(void *buf, dword size);
};

/*
 *	CLASS ht_mem_stream
 */

class ht_mem_stream: public ht_stream {
protected:
	STREAMOFS ofs;
	dword pos;
	dword bufsize, dsize;
	byte *buf;

	void extendbuf();
public:
			void	init();
			void	init(STREAMOFS ofs, dword size);
	virtual	void done();
/* overwritten */
	virtual char *get_name();
	virtual dword get_size();
	virtual dword read(void *buf, dword size);
	virtual int seek(STREAMOFS offset);
	virtual STREAMOFS tell();
	virtual dword write(void *buf, dword size);
};

/*
 *	CLASS ht_cache_page
 */

class ht_cache_page: public ht_data {
public:
	dword read_count;
	dword write_count;
	dword alloc_index;
	STREAMOFS offset;
	dword size;
	byte *data;

	ht_cache_page(STREAMOFS offset, dword size, dword alloc_index);
	~ht_cache_page();
/* new */
	virtual int read(STREAMOFS offset, byte *buf, dword len);
	virtual int writebyte(byte b, STREAMOFS offset);
	virtual int readbyte(byte *b, STREAMOFS offset);
};

/*
 *	CLASS ht_cached_stream
 */

/* CACHED_STREAM_CACHE_READS
	Allow read-caching, memory used is limited by maxcachesize  */
#define CACHED_STREAM_CACHE_READS			1

/* CACHED_STREAM_CACHE_WRITES
	Allow write-caching, memory used is limited by maxcachesize (exception see below) */
#define CACHED_STREAM_CACHE_WRITES			2

/* CACHED_STREAM_ONLY_EXPLICIT_FLUSHES
	Disables memory limit "maxcachesize" for write access !
	You MUST include CACHED_STREAM_CACHE_WRITES if specifing this flag ! */
#define CACHED_STREAM_ONLY_EXPLICIT_FLUSHES	4

class ht_dtree;

class ht_cached_stream: public ht_layer_stream {
protected:
	ht_dtree *cache;
	dword maxcachesize;
	dword flags;
	dword alloc_index;
	STREAMOFS offset;
	dword size;

	dword read_pages;
	dword write_pages;

	dword page_granularity;
	dword page_mask;
/* new */
	virtual ht_cache_page *cache_page_create(STREAMOFS offset);
	virtual void cache_page_destroy(ht_cache_page *c);
	virtual ht_cache_page *cache_page_find(STREAMOFS offset);
	virtual void cache_page_flush(ht_cache_page *c);

	void flush_and_free_least_used_page();
	void free_read_pages();
	int writebyte(STREAMOFS offset, byte b);
	int readbyte(STREAMOFS offset, byte *b);
	ht_cache_page *read_miss(STREAMOFS offset);
public:
			void	init(ht_stream *stream, int own_stream, dword maxcachesize, int flags);
			void	init(ht_stream *stream, int own_stream, dword maxcachesize, int flags, dword page_granularity);
	virtual	void done();
/* overwritten */
	void flush();
	virtual dword get_size();
	virtual void invalidate();
	virtual dword read(void *buf, dword size);
	virtual int seek(STREAMOFS offset);
	virtual void set_mode(dword mode);
	virtual STREAMOFS tell();
	virtual dword write(void *buf, dword size);
};

/*
 *	CLASS ht_cache_page_pbd
 */

/* pbd = per-bit write dirtyness */

class ht_cache_page_pbd: public ht_cache_page {
public:
	byte *writedirty;
	ht_stream *stream;

	ht_cache_page_pbd(STREAMOFS offset, dword size, dword alloc_index, ht_stream *stream);
	~ht_cache_page_pbd();
/* new */
	int isdirty(STREAMOFS offset, int range);
	void cleanup();
};

/*
 *	CLASS ht_cached_stream_pbd
 */

class ht_cached_stream_pbd: public ht_cached_stream {
protected:
/* overwritten */
	virtual ht_cache_page *cache_page_create(STREAMOFS offset);
	virtual void cache_page_flush(ht_cache_page *c);
public:
			void	init(ht_stream *stream, int own_stream, dword maxcachesize, int flags);
			void	init(ht_stream *stream, int own_stream, dword maxcachesize, int flags, dword page_granularity);
/* overwritten */
	virtual int isdirty(STREAMOFS offset, int range);
};

/*
 *	string stream functions
 */

/* 0-terminated ASCII */
char *getstrz(ht_stream *stream);
void putstrz(ht_stream *stream, char *str);

/* pascal-style ASCII */
char *getstrp(ht_stream *stream);
void putstrp(ht_stream *stream, char *str);

/* 0-terminated wide/unicode */
char *getstrw(ht_stream *stream);
void putstrw(ht_stream *stream, char *str);

/*
 *   file system functions
 */

int canonicalize(char *in_file, char *out_file);

#endif /* !__HTSTREAM_H__ */
