/*
 *      HT Editor
 *      store.cc
 *
 *      Copyright (C) 1999, 2000, 2001 Sebastian Biallas (sb@web-productions.de)
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2 as
 *      published by the Free Software Foundation.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include "store.h"
#include "htdebug.h"

char hexchars2[17]="0123456789abcdef";

/*
 *      store_filter
 */
void ht_stream_filter::init(ht_stream *s)
{
	assert(s);
	stream = s;
}

void ht_stream_filter::done()
{
}

void *ht_stream_filter::get_binary(int size, char *desc)
{
	// ABSTRACT
	return NULL;
}

bool ht_stream_filter::get_bool(char *desc)
{
	// ABSTRACT
	return false;
}

int  ht_stream_filter::get_int_dec(int size, char *desc)
{
	// ABSTRACT
	return 0;
}

int  ht_stream_filter::get_int_hex(int size, char *desc)
{
	// ABSTRACT
	return 0;
}

object *ht_stream_filter::get_object(char *name)
{
	// ABSTRACT
	return NULL;
}

void ht_stream_filter::get_seperator()
{
	// ABSTRACT
}

char *ht_stream_filter::get_string(char *desc)
{
	// ABSTRACT
	return NULL;
}

void ht_stream_filter::put_binary(void *mem, int size, char *desc)
{
	// ABSTRACT
}

void ht_stream_filter::put_bool(bool b, char *desc)
{
	// ABSTRACT
}

void ht_stream_filter::put_info(char *info)
{
	// ABSTRACT
}

void ht_stream_filter::put_int_dec(int a, int size, char *desc)
{
	// ABSTRACT
}

void ht_stream_filter::put_int_hex(int a, int size, char *desc)
{
	// ABSTRACT
}

void ht_stream_filter::put_object(object *obj, char *name)
{
	// ABSTRACT
}

void ht_stream_filter::put_seperator()
{
	// ABSTRACT
}

void ht_stream_filter::put_string(char *string, char *desc)
{
	// ABSTRACT
}

/*
 *   bin_filter
 */
void *bin_filter::get_binary(int size, char *desc)
{
	void *a = malloc(size);
	stream->read(a, size);
	return a;
}

bool    bin_filter::get_bool(char *desc)
{
	bool b=false;
	stream->read(&b, 1);
	return b;
}

int  bin_filter::get_int_dec(int size, char *desc)
{
	int a;
	stream->read(&a, size);
	return a;
}

int  bin_filter::get_int_hex(int size, char *desc)
{
	int a;
	stream->read(&a, size);
	return a;
}

object *bin_filter::get_object(char *name)
{
	// moep
	return NULL;
}

void bin_filter::get_seperator()
{
	// empty
}

char *bin_filter::get_string(char *desc)
{
	return getstrz(stream);
}

void bin_filter::put_binary(void *mem, int size, char *desc)
{
	stream->write(mem, size);
}

void    bin_filter::put_bool(bool b, char *desc)
{
	b = (b)?1:0;
	stream->write(&b, 1);
}

void bin_filter::put_info(char *info)
{
	// empty
}

void bin_filter::put_int_dec(int a, int size, char *desc)
{
	stream->write(&a, size);
}

void bin_filter::put_int_hex(int a, int size, char *desc)
{
	stream->write(&a, size);
}

void bin_filter::put_object(object *obj, char *name)
{
	obj->store(this);
}

void bin_filter::put_seperator()
{
	// empty
}

void bin_filter::put_string(char *string, char *desc)
{
	stream->write(string, strlen(string)+1);
}


/*
 *   txt_filter
 */
void txt_filter::init(ht_stream *s)
{
	ht_stream_filter::init(s);
	indent = 0;
}

void txt_filter::done()
{
	ht_stream_filter::done();
}

void *txt_filter::get_binary(int size, char *desc)
{
	return NULL;
}

bool    txt_filter::get_bool(char *desc)
{
	return false;
}

int  txt_filter::get_int_dec(int size, char *desc)
{
	return 0;
}

int  txt_filter::get_int_hex(int size, char *desc)
{
	return 0;
}

object *txt_filter::get_object(char *name)
{
	return NULL;
}

void txt_filter::get_seperator()
{
}

char *txt_filter::get_string(char *desc)
{
	return NULL;
}

void txt_filter::put_binary(void *mem, int size, char *desc)
{
	put_s(desc);
	put_s(": \"");
	for(int i=0; i<size; i++) {
		byte a = *((byte *)mem+i);
		put_char(hexchars2[(a & 0xf0) >> 4]);
		put_char(hexchars2[(a & 0x0f)]);
		if (i+1<size) putchar(' ');
	}
	put_s("\"\n");
}

void    txt_filter::put_bool(bool b, char *desc)
{
	put_s(desc);
	put_s(": ");
	if (b) put_s("true"); else put_s("false");
	put_char('\n');
}

void txt_filter::put_info(char *info)
{
	put_s("# ");
	put_s(info);
	put_char('\n');
}

void txt_filter::put_int_dec(int a, int size, char *desc)
{
	put_s(desc);
	put_s(": ");
	int b;
	switch (size) {
		case 1:
			b = (char) a;
		case 2:
			b = (short) a;
		case 4:
			b = (int) a;
		default:
			b = a;
	}
	char number[12];
	sprintf(number, "%d\n", a);
	puts(number);
}

void txt_filter::put_int_hex(int a, int size, char *desc)
{
	put_s(desc);
	put_s(": ");
	int b;
	switch (size) {
		case 1:
			b = (char) a;
		case 2:
			b = (short) a;
		case 4:
			b = (int) a;
		default:
			b = a;
	}
	char number[12];
	char number2[12];
	sprintf(number, "0x%%0%dx\n", size*2);
	sprintf(number2, number, a);
	put_s(number2);
}

void txt_filter::put_object(object *obj, char *name)
{
	put_s("object: `");
	put_s(name);
	put_s("'\n");
	indent++;
	obj->store(this);
	indent--;
	put_s("end of `");
	put_s(name);
	put_s("'\n");
}

void txt_filter::put_seperator()
{
	put_s("# ------------------------ \n");
}

void txt_filter::put_string(char *string, char *desc)
{
}

void txt_filter::put_s(char *s)
{
	stream->write(s, strlen(s));
}

void txt_filter::put_char(char c)
{
	stream->write(&c, 1);
}

